/*========================================================================
                              
                                ** UMOP **
		** Universal Multi-Agent OBDD-based Planner **

            Copyright (C) 1999 by Rune M. Jensen and Manuela M. Veloso
            Computer Science Department, Carnegie Mellon University.
            
    Permission is hereby granted to distribute the UMOP software for 
    research purposes, provided this notice is included with any such 
    distribution. 

      *** WE ARE NOT RESPONSIBLE FOR ANY KIND OF DAMAGE TO YOUR  ***
      *** FILES, DATA, HARDWARE, LOSS OF MONEY, SYSTEM CRASHES   *** 
      *** OR ANY OTHER THING YOU MIGHT COME UP WITH.             ***
      *** - USE THIS PROGRAM OF YOUR OWN FREE WILL !!!           ***

========================================================================*/
/****************************************************** 
 * File  : bddprint.cc   
 * Desc. : functions for nice prints of BDD trans. rel.
 *         of NADL domains.
 * Author: Rune M. Jensen, CS, CMU, (IAU, DTU)
 * Date  : 3/4/99
 ******************************************************/

#include <iostream.h>
#include <fstream.h>
#include <stream.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <set>
#include <map>
#include <string>
#include <vector>
#include "fsm.hpp"
#include "bddprint.hpp"
#include "common.hpp"


bddprintabs::bddprintabs(char *filename,domdef *d) { 
  strcpy(outfilename,filename); 
  outfile.open(filename,ios::out);
  def = d; 
} 

void bddprintabs::reset() {
  outfile.close();
  outfile.open(outfilename,ios::out);
}




// reads one variable assignment
// of the bdd_printset format 
int bddprintabs::parse(int *as,int aslen) {
  int i, varno, ordno,value;
  char c;
  
  //check if file T or False
  c = getc(setfile);
  
  if ( c == 'F' || c==EOF )
    return 0;
  else 
    if ( c == 'T' )
      {
	// set all variables to -1 (any value)
	for (i=0;i<aslen;i++)
	  as[i] = -1;
	return 1;
      }
    else
      if ( c == '<' ) 
	{  
	  // there is an assignment to read

	  // set all variables to -1 (any value)
	  for (i=0;i<aslen;i++)
	    as[i] = -1;
	  
	  // read assignment 	  
	  while(c != '>') {
	    
	    // pre: pointer is at first digit
	    // in var. assignment
	    if (fscanf(setfile,"%d/%d:%d",&varno,&ordno,&value) != 3)
	      {
		cout << "\nbddprint.cc printbddabs::parse: scan error";
		cout << "\nexiting\n\n";
	      }
	    as[varno] = value;
	    // post: pointer is just after last difit
	    // in var. assignment 
	    c = getc(setfile);
	  }
	  return 1;
	}
}

	
  

// aux. for printbdd 



void printacts(vector<action> act, ofstream &out,int *actlen) {
  for (int i = 0; i < act.size(); i++)
    {
      out << act[i].name << "(" << i << ") ";
      if (act[i].name.length() + 2 > *actlen) *actlen = act[i].name.length() + 2;
    }
}



void printagts(vector<agtdef> agt,ofstream &out,int *actlen) {

  for (int i = 0; i < agt.size(); i++)
    {
      out << agt[i].name << "(" << i << ","; 
      if (agt[i].sclass == sc_env)
	out << "env): ";
      else
	out << "sys): ";
      actlen[i] = agt[i].name.length() + 2;
      printacts(agt[i].act,out,&actlen[i]);
      out << "\n";
      
    }
}




void mkspace(ofstream &out,int n) {
  int i;
  for (i=0;i<n;i++)
    out << " ";
}

int max(int a,int b) {
  return a > b ? a : b;
} 

int bin2intstr(int *as,int *var,int length,char *buf) {
  int i,exp,res;
  
  res = 0;
  exp = length - 1;
  for (i= 0; i < length; i++)
    if (as[var[i]] < 0) 
      return 0;
    else
      res += as[var[i]]*int(pow(2,exp--));
    sprintf(buf,"%d",res);
    return 1;
}
  

char *rawrep(int *as,int *var,int length,char *buf) {
  int i;
 
  strcpy(buf,"");
  for (i=0; i<length; i++)
    switch (as[var[i]]) {
    case 0 : 
      strcat(buf,"0");
      break;
    case 1 :
      strcat(buf,"1");
      break;
    default:
      strcat(buf,"*");
      break;
    }
  return buf;
}
    



void bddprintabs::bddprint(bdd b) {
   
  int        *as;
  int         i,k,len,plen,aslen;
  char        buf[MAXNAMELENGTH];         


  // open outputfile for bdd_printset
  if ( (setfile = fopen("deleteme.txt","w+")) == NULL)
    {
      cout << "bddprint.cc printbddabs::bddprintline: ";
      cout << "cannot open \"deleteme.txt\"\nexiting\n\n";
      exit(1);
    }

  // make the set
  bdd_fprintset(setfile,b);
  fflush(setfile);
  rewind(setfile);
  
  // print action description (and fill the action length table)
  actlen = new int[def->agt.size()];
  printagts(def->agt,outfile,actlen);
  outfile << "\n";


  // write header
  // agt names
  for (i=0; i < def->agt.size(); i++) {
    outfile << def->agt[i].name;
    mkspace(outfile,actlen[i] - def->agt[i].name.length());
  }
  
  // write header 
  // write all old state variables

    for (k = 0; k < def->varnames.size(); k++)
      {
	vardef var(def->varmap[def->varnames[k]]);
	len = max(var.name.length() + 2, (var.length + 1)*2);
	outfile << var.name;
	mkspace(outfile,len - var.name.length());
      }
    outfile << "\n";

  // Init assignment table
  aslen = def->bddvarnum;
  as = new int[aslen]; 

  // go through all lines in the table
  while (parse(as,aslen))
    {
      // write action var info
      for (i=0; i<def->agt.size(); i++)
	{
	  char str[MAXNAMELENGTH];
	  sprintf(str,"%d",i);
	  vardef var(def->varmap[str]);        
	  if (bin2intstr(as,var.var,var.length,buf))
	    {
	      if ( atoi(buf) < def->agt[i].act.size() )
		{
		  // action can be written by its name
		  outfile << (def->agt[i].act[atoi(buf)].name);
		  mkspace(outfile,actlen[i] - def->agt[i].act[atoi(buf)].name.length());
		}
	      else
		{
		  // action doesn't exist write the number
		  outfile << buf;
		  mkspace(outfile,actlen[i] - strlen(buf));
		}
	    }
	  else
	    {
	      // the line defines a set of actions
	      // we have to write the number pattern
	      outfile << rawrep(as,var.var,var.length,buf);
	      mkspace(outfile,actlen[i] - strlen(buf));
	    }
	}
      
      // write individual variable info
      for (k = 0; k < def->varnames.size(); k++)
	{
	  vardef var(def->varmap[def->varnames[k]]);   // old var 
	  vardef varn(def->varmap[def->varnames[k] + "'"]); // new var

	  len = max(var.name.length() + 2, (var.length + 1)*2);
	  plen = 0;
    	  if (var.type == vt_bool)
	    {
	      // write old state var
	      outfile << rawrep(as,var.var,var.length,buf) << " ";
	      // write new state var
	      outfile << rawrep(as,varn.var,varn.length,buf);
	      mkspace(outfile,len - 3);
	    }
	  else
	    {
	      // variable is scalar
	      // write old state scalar variable
	      if (bin2intstr(as,var.var,var.length,buf))
		{
		  // scalar can be represented as a number
		  outfile << buf << " ";
		  plen += strlen(buf) + 1;
		}
	      else
		{
		  // scalar must be represented by its number pattern
		  outfile << rawrep(as,var.var,var.length,buf) << " ";
		  plen += strlen(buf) + 1;
		}


	      // write new state scalar variable
	      if (bin2intstr(as,varn.var,varn.length,buf))
		{
		  // scalar can be represented as a number
		  outfile << buf << " ";
		  plen += strlen(buf) + 1;
		}
	      else
		{
		  // scalar must be represented by its number pattern
		  outfile << rawrep(as,varn.var,varn.length,buf) << " ";
		  plen += strlen(buf) + 1;
		}
	      mkspace(outfile,len - plen);
	    }
	}
      // next line
      outfile << "\n";
    }
  outfile.flush();
  
  
  // remove outputfile for bdd_printset
  system("rm -f deleteme.txt");
  fclose(setfile);
}
	      
	    
void bddprintabs::bddinfo(bdd b) {
  //write BDD size information
  cout << "\n\nBDD node count    = " << bdd_nodecount(b);
  cout << "\nlog(BDD satcount) = " << bdd_satcountln(b) << "\n\n";
}
      

void bddprintabs::bddprinthead(){
   
  int         len,i,k;
  
  // print action description (and fill the action length table)
  actlen = new int[def->agt.size()];
  printagts(def->agt,outfile,actlen);
  outfile << "\n";

  // write header
  // agt names
  for (i=0; i < def->agt.size(); i++) {
    outfile << def->agt[i].name;
    mkspace(outfile,actlen[i] - def->agt[i].name.length());
  }
  
  // write header 
  // write all old state variables
  for (k = 0; k < def->varnames.size(); k++)
    {
      vardef  var(def->varmap[def->varnames[k]]);
      len = max(var.name.length() + 2, (var.length + 1)*2);
      outfile << var.name;
      mkspace(outfile,len - var.name.length());
  }
  outfile << "\n";
}
	    
  



