/*========================================================================
                              
                                ** UMOP **
		** Universal Multi-Agent OBDD-based Planner **

            Copyright (C) 1999 by Rune M. Jensen and Manuela M. Veloso
            Computer Science Department, Carnegie Mellon University.
            
    Permission is hereby granted to distribute the UMOP software for 
    research purposes, provided this notice is included with any such 
    distribution. 

      *** WE ARE NOT RESPONSIBLE FOR ANY KIND OF DAMAGE TO YOUR  ***
      *** FILES, DATA, HARDWARE, LOSS OF MONEY, SYSTEM CRASHES   *** 
      *** OR ANY OTHER THING YOU MIGHT COME UP WITH.             ***
      *** - USE THIS PROGRAM OF YOUR OWN FREE WILL !!!           ***

========================================================================*/


/****************************************************** 
 * File  : dissets.cc  
 * Desc. : Implementation of a disjoint set ADT,
 *         using both union by size and path compression.
 *         See "Algorithms and Data Structure, Design, 
 *         Correctness, Analysis", Jeffery H. Kingston. 
 * Author: Rune M. Jensen
 * Date  : 3/16/99
 ******************************************************/

#include <string.h>
#include <stream.h>
#include "common.hpp"
#include "dissets.hpp"

  
void disset::makeset(string id) {
  disnode *n;

  // add new disnode  node
  n = new disnode;
  n->parent = NULL;
  n->size = 1;

  // add node in idmap
  idmap[id] = n;
}
 
// IN
//  id : variable identifier
// OUT
//  disnode : identifier for id's set
disnode *disset::find(string id) {
  disnode *x,*y,*z,*tmp;

  x = idmap[id];
  if (x == NULL)
    {
      cout << "dissets.cc disset::find : cannot find id \"" << id;
      cout << "\" in map \nexiting\n\n";
      exit(1);
    }

  y = x;
  // find set id.
  while (y->parent)
    y = y->parent;
  
  // compress the path
  z = x;
  while (z->parent)
    {
      tmp = z->parent;
      z->parent = y;
      z = tmp;
    }
  return y;
}


disnode *disset::setunion(disnode *v, disnode *w) {
  if (v->size < w->size)
    {
      v->parent = w;
      w->size += v->size;
      return w;
    }
  else
    {
      w->parent = v;
      v->size += w->size;
      return v;
    }
}


// pre: all unions made on disset 
void disset::mkdisjointsets() {
  int i;
  disnode *setrep;

  i = 0;
  for (map<string, disnode*>::iterator m = idmap.begin(); m != idmap.end(); ++m)
      if (m->second->parent == NULL)
	m->second->setnum = i++;

  subsetnum = i;

  vector< set<string> >  d(subsetnum);

  disjointset = d;
  for (map<string,disnode*>::iterator m = idmap.begin(); m != idmap.end(); ++m)
    {
      setrep = find(m->first);
      disjointset[setrep->setnum].insert(m->first);
    }
}
