%token SYSTEM ENVIRONMENT INITIALLY GOAL 
%token PRE EFF VAR AGT SCALAR ARROW TRUE FALSE
%token NE ID NUMBER VARIABLES BOOL
%token BIIMPL IMPL AND OR 

%nonassoc ARROW
%left BIIMPL IMPL
%left OR
%left AND
%nonassoc '~' 
%left '+' '-'

%start problem
%%
/****************************************************** 
 * File  : nadl.y
 * Desc. : Yacc file for NADL language
 *         Abstract Syntax described in domain.h
 *         See also lex file nadl.l
 * Author: Rune M. Jensen, CS, CMU (IAU, DTU)
 * Date  : 3/2/99
 ******************************************************/

problem:     VARIABLES vardecls SYSTEM agtdecls ENVIRONMENT agtdecls INITIALLY formula GOAL formula  
                    { prob = new problem((list<vardecl>*) $2, 
					 (list<agent>*)   $4,
					 (list<agent>*)   $6,
					 (formula*) $8,
					 (formula*) $10 );
		    }              
                ;

vardecls:       
                    { $$ = (int) new list<vardecl>; } 
              | varlst vardecls
	            { vardec = (vardecl*) $1;
		      vars = (list<vardecl>*) $2;
		      vars->push_front(*vardec); 
		      $$ = (int) vars; }
                ;

varlst:         vartype idlst
	            { vardec = new vardecl((vartype*) $1, (set<string>*) $2);
		      $$ = (int) vardec; }                
                ;

vartype:        BOOL 
	            { vatype = new vartype(vt_bool,0);
		      $$ = (int) vatype; }                
              | SCALAR '(' number ')'
	            { vatype = new vartype(vt_scalar,(int) $3);
		      $$ = (int) vatype; }                
                ;

idlst:        
	            { $$ = (int) new set<string>; }
              | idlst1
		    { $$ = $1; }  
                ;

idlst1:         id
                    { ids = new set<string>;
                      ids->insert((char*) $1);
		      $$ = (int) ids; }          
              | id ',' idlst1
                    { ids = (set<string>*) $3;
                      ids->insert((char*) $1);
		      $$ = (int) ids; }          
                ;

              
agtdecls: 
	            { $$ = (int) new list<agent>; }  
              | AGT id actiondecls agtdecls
		    { agtdec = (list<agent>*) $4;
		      agt = new agent((char*) $2, (list<action>*) $3);
                      agtdec->push_front(*agt);
                      $$ = (int) agtdec; }
                ;

actiondecls:    actiondecl
                    { actdec = new list<action>;
		      act = (action*) $1;
		      actdec->push_front(*act);		      
		      $$ = (int) actdec; }
              | actiondecl actiondecls
		    { actdec = (list<action>*) $2;
		      act = (action*) $1;
                      actdec->push_front(*act);
                      $$ = (int) actdec; } 
                ;

actiondecl:     id VAR idlst PRE formula EFF formula
                    { act = new action( (char*) $1,(set<string>*) $3,(formula*) $5,(formula*) $7);
                      $$ = (int) act; }
                ;

formula:        formula1 ARROW formula1 ',' formula1
                    { $$ = (int) new formula(ft_ite,NULL,(formula*) $1,(formula*) $3,(formula*) $5); }
              | formula1
                    { $$ = $1; } 
                ;

formula1:       formula1 IMPL formula1
                    { $$ = (int) new formula(ft_impl,NULL,(formula*) $1,(formula*) $3,NULL); }
              | formula1 BIIMPL formula1
                    { $$ = (int) new formula(ft_biimpl,NULL,(formula*) $1,(formula*) $3,NULL); }
              | formula1 OR formula1
                    { $$ = (int) new formula(ft_or,NULL,(formula*) $1,(formula*) $3,NULL); }
              | formula1 AND formula1
                    { $$ = (int) new formula(ft_and,NULL,(formula*) $1,(formula*) $3,NULL); }
              | '~' formula1
                    { $$ = (int) new formula(ft_neg,NULL,(formula*) $2,NULL,NULL); }
              | '(' formula ')'
                    { $$ = (int) new formula(ft_paren,NULL,(formula*) $2,NULL,NULL); }
              | TRUE
                    { $$ = (int) new formula(ft_true,NULL,NULL,NULL,NULL); }
              | FALSE
                    { $$ = (int) new formula(ft_false,NULL,NULL,NULL,NULL); }
              | atom
                    { $$ = (int) new formula(ft_atom,(atom*) $1,NULL,NULL,NULL); }
                ;

atom:           id 
                    { $$ = (int) new atom(at_boolvar,NULL,(char*) $1); }
              | numprop
		    { $$ = (int) new atom(at_numprop,(numberprop*) $1,""); }
                ;

numprop:        numberexp relop numberexp 
                    { $$ = (int) new numberprop((relop) $2,(numberexp*) $1,(numberexp*) $3); }
                ;

relop:          '=' 
                    { $$ = (int) ro_eq; } 
              | NE 
                    { $$ = (int) ro_ne; } 
              | '>' 
                    { $$ = (int) ro_gt; } 
              | '<'
                    { $$ = (int) ro_lt; } 
                ;

numberexp:      id
                    { $$ = (int) new numberexp(nt_var,0,NULL,NULL,(char*) $1); }
              | number
                    { $$ = (int) new numberexp(nt_number,(int) $1,NULL,NULL,""); }
              | numberexp '+' numberexp
                    { $$ = (int) new numberexp(nt_plus,0,(numberexp*) $1,(numberexp*) $3,""); }
              | numberexp '-' numberexp 
                    { $$ = (int) new numberexp(nt_minus,0,(numberexp*) $1,(numberexp*) $3,""); }
                ;       

id:             ID
                    { s = (char*) malloc(MAXNAMELENGTH);
                      strcpy(s,yytext);
                      $$ = (int) s; }
                ;

number:         NUMBER
                    { $$ = (int) atoi(yytext);  }
                ;

%%
#include <stdio.h>
#include <stdlib.h>
#include <list>
#include <string>
#include <set>
#include "domain.hpp"

#define MAXNAMELENGTH 128

extern char *yytext;  /* defined in nadl.l */
extern int lineno;    /* defined in nadl.l */

problem        *prob;  /* pointer to abstract syntax (only extern var)*/

list<agent>    *agtdec;
agent          *agt;
list<vardecl>  *vars;
vardecl        *vardec;
vartype        *vatype;
set<string>    *ids;
char           *s;
list<action>   *actdec;
action         *act;       
atom           *a;


/* prototypes */

int yylex(void);

void yyerror(char *s) {
  fflush(stdout);
  printf("\n%s! lineno:%d at:\"%s\"\n", s, lineno, yytext);
}





